"""
This module contains `Recorder` subclasses useful in the evaluation of model
 performance and calibration.

Several different evaluation metrics are implemented in this module. Many are
 well known to modellers. However an important reference for this implementation
 is `[1]_` which reviews the available metrics for watershed models.

..  [1] Moriasi, D.N., et al. (2007) Model Evaluation Guidelines for Systematic
    Quantification of Accuracy in Watershed Simulations. Transactions of the ASABE, 50, 885-900.
    http://dx.doi.org/10.13031/2013.23153


"""
from ._recorders import NumpyArrayNodeRecorder
import numpy as np


class AbstractComparisonNodeRecorder(NumpyArrayNodeRecorder):
    """Base class for all Recorders performing timeseries comparison of `Node` flows"""

    def __init__(self, model, node, observed, **kwargs):
        super(AbstractComparisonNodeRecorder, self).__init__(model, node, **kwargs)
        self.observed = observed
        self._aligned_observed = None

    def setup(self):
        super(AbstractComparisonNodeRecorder, self).setup()
        # Align the observed data to the model
        from pywr.parameters import align_and_resample_dataframe

        self._aligned_observed = align_and_resample_dataframe(
            self.observed, self.model.timestepper.datetime_index
        )


class RootMeanSquaredErrorNodeRecorder(AbstractComparisonNodeRecorder):
    """Recorder evaluates the RMSE between model and observed"""

    def values(self):
        mod = self.data
        obs = self._aligned_observed
        return np.sqrt(np.mean((obs - mod) ** 2, axis=0))


class MeanAbsoluteErrorNodeRecorder(AbstractComparisonNodeRecorder):
    """Recorder evaluates the MAE between model and observed"""

    def values(self):
        mod = self.data
        obs = self._aligned_observed
        return np.mean(np.abs(obs - mod), axis=0)


class MeanSquareErrorNodeRecorder(AbstractComparisonNodeRecorder):
    """Recorder evaluates the MSE between model and observed"""

    def values(self):
        mod = self.data
        obs = self._aligned_observed
        return np.mean((obs - mod) ** 2, axis=0)


class PercentBiasNodeRecorder(AbstractComparisonNodeRecorder):
    """Recorder evaluates the percent bias between model and observed"""

    def values(self):
        mod = self.data
        obs = self._aligned_observed
        return np.sum(obs - mod, axis=0) * 100 / np.sum(obs, axis=0)


class RMSEStandardDeviationRatioNodeRecorder(AbstractComparisonNodeRecorder):
    """Recorder evaluates the RMSE-observations standard deviation ratio between model and observed"""

    def values(self):
        mod = self.data
        obs = self._aligned_observed
        return np.sqrt(np.mean((obs - mod) ** 2, axis=0)) / np.std(obs, axis=0)


class NashSutcliffeEfficiencyNodeRecorder(AbstractComparisonNodeRecorder):
    """Recorder evaluates the Nash-Sutcliffe efficiency model and observed"""

    def values(self):
        mod = self.data
        obs = self._aligned_observed
        obs_mean = np.mean(obs, axis=0)
        return 1.0 - np.sum((obs - mod) ** 2, axis=0) / np.sum(
            (obs - obs_mean) ** 2, axis=0
        )



"""
This module contains a set of pywr._core.Parameter subclasses for defining control curve based parameters.
"""
from ._control_curves import (
    ControlCurveParameter,
    BaseControlCurveParameter,
    ControlCurveInterpolatedParameter,
    ControlCurveIndexParameter,
    ControlCurvePiecewiseInterpolatedParameter,
)



import numpy as np
from . import BaseOptimisationWrapper

import logging

logger = logging.getLogger(__name__)


class PygmoWrapper(BaseOptimisationWrapper):
    def fitness(self, solution):
        logger.info("Evaluating solution ...")

        for ivar, var in enumerate(self.model_variables):
            j = slice(self.model_variable_map[ivar], self.model_variable_map[ivar + 1])
            var.set_double_variables(np.array(solution[j]).copy())

        self.model.reset()
        self.run_stats = self.model.run()

        objectives = []
        for r in self.model_objectives:
            sign = 1.0 if r.is_objective == "minimise" else -1.0
            value = r.aggregated_value()
            objectives.append(sign * value)

        # Return separate lists for equality and inequality constraints.
        # pygmo requires that inequality constraints are all of the form g(x) <= 0
        # Therefore these are converted to this form from their respective bounds.
        eq_constraints = []
        ineq_constraints = []
        for r in self.model_constraints:
            x = r.aggregated_value()
            if r.is_double_bounded_constraint:
                # Need to create two constraints
                ineq_constraints.append(r.constraint_lower_bounds - x)
                ineq_constraints.append(x - r.constraint_upper_bounds)
            elif r.is_equality_constraint:
                eq_constraints.append(x)
            elif r.is_lower_bounded_constraint:
                ineq_constraints.append(r.constraint_lower_bounds - x)
            elif r.is_upper_bounded_constraint:
                ineq_constraints.append(x - r.constraint_upper_bounds)
            else:
                raise RuntimeError(
                    f'The bounds if constraint "{r.name}" could not be identified correctly.'
                )

        # Return values to the solution
        logger.info(
            f"Evaluation completed in {self.run_stats.time_taken:.2f} seconds "
            f"({self.run_stats.speed:.2f} ts/s)."
        )
        return objectives + eq_constraints + ineq_constraints

    def get_bounds(self):
        """Return the variable bounds."""
        lower = []
        upper = []
        for var in self.model_variables:

            if var.double_size > 0:
                lower.append(var.get_double_lower_bounds())
                upper.append(var.get_double_upper_bounds())

            if var.integer_size > 0:
                lower.append(var.get_integer_lower_bounds())
                upper.append(var.get_integer_upper_bounds())

        lower = np.concatenate(lower)
        upper = np.concatenate(upper)

        if len(lower) != len(upper):
            raise ValueError(
                "Upper and lower bounds are different lengths. Malformed bound data from Parameter:"
                ' "{}"'.format(var.name)
            )

        return lower, upper

    def get_nobj(self):
        return len(self.model_objectives)

    def get_nec(self):
        return len([c for c in self.model_constraints if c.is_equality_constraint])

    def get_nic(self):
        count = 0
        for c in self.model_constraints:
            if c.is_double_bounded_constraint:
                count += 2
            elif c.is_lower_bounded_constraint or c.is_upper_bounded_constraint:
                count += 1
        return count



import numpy as np
import platypus
from . import (
    cache_constraints,
    cache_objectives,
    cache_variable_parameters,
    BaseOptimisationWrapper,
)

import logging

logger = logging.getLogger(__name__)


def count_constraints(constraints):
    """Count the number of constraints.

    Recorders that are doubled bounded will create two constraints in the platypus problem.
    """
    count = 0
    for c in constraints:
        if c.is_double_bounded_constraint:
            count += 2
        elif c.is_constraint:
            count += 1
        else:
            raise ValueError(f'Constraint "{c.name}" has no bounds defined.')
    return count


class PlatypusWrapper(BaseOptimisationWrapper):
    """A helper class for running pywr optimisations with platypus."""

    def __init__(self, *args, **kwargs):
        super(PlatypusWrapper, self).__init__(*args, **kwargs)

        # To determine the number of variables, etc
        m = self.model

        # Cache the variables, objectives and constraints
        variables, variable_map = cache_variable_parameters(m)
        objectives = cache_objectives(m)
        constraints = cache_constraints(m)

        if len(variables) < 1:
            raise ValueError("At least one variable must be defined.")

        if len(objectives) < 1:
            raise ValueError("At least one objective must be defined.")

        self.problem = platypus.Problem(
            variable_map[-1], len(objectives), count_constraints(constraints)
        )
        self.problem.function = self.evaluate
        self.problem.wrapper = self

        # Setup the problem; subclasses can change this behaviour
        self._make_variables(variables)
        self._make_constraints(constraints)

    def _make_variables(self, variables):
        """Setup the variable types."""

        ix = 0
        for var in variables:
            if var.double_size > 0:
                lower = var.get_double_lower_bounds()
                upper = var.get_double_upper_bounds()
                for i in range(var.double_size):
                    self.problem.types[ix] = platypus.Real(lower[i], upper[i])
                    ix += 1

            if var.integer_size > 0:
                lower = var.get_integer_lower_bounds()
                upper = var.get_integer_upper_bounds()
                for i in range(var.integer_size):
                    # Integers are cast to real
                    self.problem.types[ix] = platypus.Real(lower[i], upper[i])
                    ix += 1

    def _make_constraints(self, constraints):
        """Setup the constraints."""

        ic = 0  # platypus constraint index
        for c in constraints:
            if c.is_double_bounded_constraint:
                # Need to create two constraints
                self.problem.constraints[ic] = platypus.Constraint(
                    ">=", value=c.constraint_lower_bounds
                )
                self.problem.constraints[ic + 1] = platypus.Constraint(
                    "<=", value=c.constraint_upper_bounds
                )
                ic += 2
            elif c.is_equality_constraint:
                self.problem.constraints[ic] = platypus.Constraint(
                    "==", value=c.constraint_lower_bounds
                )
                ic += 1
            elif c.is_lower_bounded_constraint:
                self.problem.constraints[ic] = platypus.Constraint(
                    ">=", value=c.constraint_lower_bounds
                )
                ic += 1
            elif c.is_upper_bounded_constraint:
                self.problem.constraints[ic] = platypus.Constraint(
                    "<=", value=c.constraint_upper_bounds
                )
                ic += 1
            else:
                raise RuntimeError(
                    f'The bounds of constraint "{c.name}" could not be identified correctly.'
                )

    def evaluate(self, solution):
        logger.info("Evaluating solution ...")

        for ivar, var in enumerate(self.model_variables):
            j = slice(self.model_variable_map[ivar], self.model_variable_map[ivar + 1])
            x = np.array(solution[j])
            assert len(x) == var.double_size + var.integer_size
            if var.double_size > 0:
                var.set_double_variables(np.array(x[: var.double_size]))

            if var.integer_size > 0:
                ints = np.round(np.array(x[-var.integer_size :])).astype(np.int32)
                var.set_integer_variables(ints)

        self.run_stats = self.model.run()

        objectives = []
        for r in self.model_objectives:
            sign = 1.0 if r.is_objective == "minimise" else -1.0
            value = r.aggregated_value()
            objectives.append(sign * value)

        constraints = []
        for c in self.model_constraints:
            x = c.aggregated_value()
            if c.is_double_bounded_constraint:
                # Double bounded recorder is translated to two platypus constraints.
                constraints.extend([x, x])
            else:
                constraints.append(x)

        # Return values to the solution
        logger.info(
            f"Evaluation completed in {self.run_stats.time_taken:.2f} seconds "
            f"({self.run_stats.speed:.2f} ts/s)."
        )
        if len(constraints) > 0:
            return objectives, constraints
        else:
            return objectives


class PywrRandomGenerator(platypus.RandomGenerator):
    """A Platypus Generator that injects current and/or alternative setups of the Pywr model into the population.

    When use_current is true the first Solution returned from the generate method is taken from the wrapper
    (i.e. the Pywr model being wrapped) as the current values of the variable Parameters. This allows the population
    to be seeded with the current model configuration, which is often an initial solution. Additional solutions
    can be provided in as an iterable of solutions. These can come from an alternative source such as previous
    optimisation.

    Parameters
    ==========
    wrapper : PlatypusWrapper
        Wrapper from which to grab the current model and decision variables.
    use_current: Bool
        Whether to generate an initial solution using the model's current configuration. Default is true.
        Set this to False and pass some solutions to use pre-generated
    solutions : List of dicts
        An iterable of initial solutions to use (default is None). If given these alternative solutions
        are provided to Platypus in order. Each item in the list should be a dictionary containing keys
        for each of the variable Parameters in the optimisation. The value of each key should be another
        dictionary container keys "doubles" and/or "integers" to provide the appropriate values as
        dictated by the Parameter's type.
    """

    def __init__(self, *args, **kwargs):
        self.wrapper = kwargs.pop("wrapper", None)
        self.use_current = kwargs.pop("use_current", True)
        self.solutions = kwargs.pop("solutions", None)
        super().__init__(*args, **kwargs)
        self._wrapped_generated = False
        self._solution_pointer = 0

    def generate(self, problem):
        solution = None
        if self.wrapper is not None:
            if self.use_current and not self._wrapped_generated:
                solution = platypus.Solution(problem)
                # Gather the variable values from the wrapper.
                variables = []
                for ivar, var in enumerate(self.wrapper.model_variables):
                    if var.double_size > 0:
                        variables.extend(
                            np.array(var.get_double_variables(), dtype=np.float64)
                        )
                    if var.integer_size > 0:
                        variables.extend(
                            np.array(var.get_integer_variables(), dtype=np.int32)
                        )
                solution.variables = variables
                self._wrapped_generated = (
                    True  # Only include one solution with the current config.
                )
            elif self.solutions is not None and self._solution_pointer < len(
                self.solutions
            ):
                # Use one of the given solutions
                solution = platypus.Solution(problem)
                given_solution = self.solutions[self._solution_pointer]
                variables = []
                for ivar, var in enumerate(self.wrapper.model_variables):
                    if var.double_size > 0:
                        variables.extend(
                            np.array(
                                given_solution[var.name]["doubles"], dtype=np.float64
                            )
                        )
                    if var.integer_size > 0:
                        variables.extend(
                            np.array(
                                given_solution[var.name]["integers"], dtype=np.int32
                            )
                        )
                solution.variables = variables
                self._solution_pointer += (
                    1  # Increment the internal pointer to return the next solution
                )

        if solution is None:
            # Default to behaviour of RandomGenerator
            solution = super().generate(problem)
        return solution



from pywr.parameters._parameters cimport IndexParameter, Parameter
from pywr.recorders._recorders cimport Recorder, StorageRecorder, NodeRecorder
from .._core cimport Timestep, Scenario, ScenarioIndex, AbstractNode, AbstractStorage, Storage

cdef class StorageThresholdRecorder(StorageRecorder):
    cdef public double threshold
    cdef int predicate
    cdef bint[:] _state

cdef class NodeThresholdRecorder(NodeRecorder):
    cdef public double threshold
    cdef int predicate
    cdef bint[:] _state



from pywr.nodes import Domain, Input, Link, Storage, PiecewiseLink, MultiSplitLink
from pywr.parameters import (
    pop_kwarg_parameter,
    ConstantParameter,
    Parameter,
    load_parameter,
)
from pywr.parameters.control_curves import ControlCurveParameter

DEFAULT_RIVER_DOMAIN = Domain(name="river", color="#33CCFF")


class RiverDomainMixin(object):
    def __init__(self, *args, **kwargs):
        # if 'domain' not in kwargs:
        #     kwargs['domain'] = DEFAULT_RIVER_DOMAIN
        if "color" not in kwargs:
            self.color = "#6ECFF6"  # blue
        super(RiverDomainMixin, self).__init__(*args, **kwargs)


class Catchment(RiverDomainMixin, Input):
    """A hydrological catchment, supplying water to the river network"""

    __parameter_attributes__ = ("cost", "flow")

    def __init__(self, *args, **kwargs):
        """Initialise a new Catchment node.

        A Catchment is an input node with a fixed inflow. I.e. min_flow and
        max_flow are the same. The value is specified as a flow keyword, and
        overrides any min_flow or max_flow keyword arguments.

        Parameters
        ----------
        flow : float or function
            The amount of water supplied by the catchment each timestep
        """
        self.color = "#82CA9D"  # green
        # Grab flow from kwargs
        flow = kwargs.pop("flow", 0.0)
        # Min/max flow set in super inits
        super(Catchment, self).__init__(*args, **kwargs)
        self.flow = flow

    def get_flow(self, timestep):
        """flow is ensured that both min_flow and max_flow are the same."""
        return self.get_min_flow(timestep)

    def __setattr__(self, name, value):
        if name == "flow":
            self.min_flow = value
            self.max_flow = value
            return
        super(Catchment, self).__setattr__(name, value)


class Reservoir(RiverDomainMixin, Storage):
    """A reservoir node with control curve.

    The Reservoir is a subclass of Storage with additional functionality to provide a
    simple control curve. The Reservoir has above_curve_cost when it is above its curve
    and the user defined cost when it is below. Typically the costs are negative
    to represent a benefit of filling the reservoir when it is below its curve.
    """

    def __init__(self, model, *args, **kwargs):
        """

        Keywords:
            control_curve - A Parameter object that can return the control curve position,
                as a percentage of fill, for the given timestep.
        """
        control_curve = pop_kwarg_parameter(kwargs, "control_curve", None)
        above_curve_cost = kwargs.pop("above_curve_cost", None)
        cost = kwargs.pop("cost", 0.0)
        if above_curve_cost is not None:
            if control_curve is None:
                # Make a default control curve at 100% capacity
                control_curve = ConstantParameter(model, 1.0)
            elif not isinstance(control_curve, Parameter):
                # Assume parameter is some kind of constant and coerce to ConstantParameter
                control_curve = ConstantParameter(model, control_curve)

            if not isinstance(cost, Parameter):
                # In the case where an above_curve_cost is given and cost is not a Parameter
                # a default cost Parameter is created.
                kwargs["cost"] = ControlCurveParameter(
                    model, self, control_curve, [above_curve_cost, cost]
                )
            else:
                raise ValueError(
                    "If an above_curve_cost is given cost must not be a Parameter."
                )
        else:
            # reinstate the given cost parameter to pass to the parent constructors
            kwargs["cost"] = cost
        super(Reservoir, self).__init__(model, *args, **kwargs)


class River(RiverDomainMixin, Link):
    """A node in the river network

    This node may have multiple upstream nodes (i.e. a confluence) but only
    one downstream node.
    """

    def __init__(self, *args, **kwargs):
        super(River, self).__init__(*args, **kwargs)


class RiverSplit(MultiSplitLink):
    """A split in the river network

    RiverSplit is a specialised version of `pywr.nodes.MultiSplitLink` with a more convenient init method.
    It is intended for a simple case of where fixed ratio of flow is required to be distributed
    to multiple downstream routes.

    Parameters
    ----------
    factors : iterable of floats
        The factors to force on the additional splits. Number of extra_slot is assumed to be one less
        than the length of factors (as per `pywr.nodes.MultiSplitLink` documentation).
    slot_names : iterable
        The identifiers to refer to the slots when connect from this Node. Length must be one more than
         the number of extra slots required.

    See also
    --------
    pywr.nodes.MultiSplitLink

    """

    def __init__(self, model, *args, nsteps=1, **kwargs):

        factors = kwargs.pop("factors")
        extra_slots = len(factors) - 1

        # These are the defaults to pass to the parent class that makes this
        # class a convenience.
        # create keyword arguments for PiecewiseLink
        costs = kwargs.pop("costs", [0.0])
        max_flows = kwargs.pop("max_flows", [None])

        super(RiverSplit, self).__init__(
            model,
            nsteps,
            *args,
            extra_slots=extra_slots,
            factors=factors,
            costs=costs,
            max_flows=max_flows,
            **kwargs,
        )


class RiverSplitWithGauge(RiverSplit):
    """A split in the river network with a minimum residual flow.

    As per `RiverSplit` but by default creates another route in the underlying object
    to model an MRF. This route is such that the MRF is not part of forced ratios. The
    intent of this object is to model the case where a proportion of flow can be
    abstracted above the MRF (e.g. 90% of flow above MRF).

    ::

                 /  -->-- X0 {max_flow: mrf,  cost: mrf_cost} -->-- \\
        A -->-- Xo  -->-- X1 {max_flow: None, cost: cost}     -->-- Xi -->-- C
                 \\  -->-- X2 {max_flow: None, cost: 0.0}      -->-- /
                           |
                           Bo -->-- Bi --> D

        Ag {nodes: [X1, X2], factors: factors}

    Parameters
    ----------
    mrf : float
        The minimum residual flow (MRF) at the gauge
    mrf_cost : float
        The cost of the route via the MRF
    cost : float
        The cost of the other (unconstrained) route
    factors : iterable of floats
        The factors to force on the additional splits. Number of extra_slot is assumed to be one less
        than the length of factors (as per `MultiSplitLink` documentation).
    slot_names : iterable
        The identifiers to refer to the slots when connect from this Node. Length must be one more than
         the number of extra slots required.
    """

    __parameter_attributes__ = ("cost", "mrf_cost", "mrf")

    def __init__(self, model, *args, mrf=0.0, cost=0.0, mrf_cost=0.0, **kwargs):
        super(RiverSplitWithGauge, self).__init__(
            model,
            *args,
            nsteps=2,
            max_flows=[mrf, None],
            costs=[mrf_cost, cost],
            **kwargs,
        )

    def mrf():
        def fget(self):
            return self.sublinks[0].max_flow

        def fset(self, value):
            self.sublinks[0].max_flow = value

        return locals()

    mrf = property(**mrf())

    def mrf_cost():
        def fget(self):
            return self.sublinks[0].cost

        def fset(self, value):
            self.sublinks[0].cost = value

        return locals()

    mrf_cost = property(**mrf_cost())

    def cost():
        def fget(self):
            return self.sublinks[1].cost

        def fset(self, value):
            self.sublinks[1].cost = value

        return locals()

    cost = property(**cost())


class Discharge(Catchment):
    """An inline discharge to the river network

    This node is similar to a catchment, but sits inline to the river network,
    rather than at the head of the river.
    """

    pass


class RiverGauge(RiverDomainMixin, PiecewiseLink):
    """A river gauging station, with a minimum residual flow (MRF)"""

    __parameter_attributes__ = ("cost", "mrf_cost", "mrf")

    def __init__(self, *args, **kwargs):
        """Initialise a new RiverGauge instance

        Parameters
        ----------
        mrf : float
            The minimum residual flow (MRF) at the gauge
        mrf_cost : float
            The cost of the route via the MRF
        cost : float
            The cost of the other (unconstrained) route
        """
        # create keyword arguments for PiecewiseLink
        cost = kwargs.pop("cost", 0.0)
        kwargs["costs"] = [kwargs.pop("mrf_cost", 0.0), cost]
        kwargs["max_flows"] = [kwargs.pop("mrf", 0.0), None]
        super(RiverGauge, self).__init__(nsteps=2, *args, **kwargs)

    def mrf():
        def fget(self):
            return self.sublinks[0].max_flow

        def fset(self, value):
            self.sublinks[0].max_flow = value

        return locals()

    mrf = property(**mrf())

    def mrf_cost():
        def fget(self):
            return self.sublinks[0].cost

        def fset(self, value):
            self.sublinks[0].cost = value

        return locals()

    mrf_cost = property(**mrf_cost())

    def cost():
        def fget(self):
            return self.sublinks[1].cost

        def fset(self, value):
            self.sublinks[1].cost = value

        return locals()

    cost = property(**cost())



from .._core cimport Timestep, Scenario, ScenarioIndex, AbstractNode, Storage, AbstractStorage
from ._parameters cimport Parameter, IndexParameter


cdef class BinaryStepParameter(Parameter):
    cdef double _value
    cdef public double output
    cdef double _lower_bounds
    cdef double _upper_bounds

cdef class RectifierParameter(Parameter):
    cdef double _value
    cdef public double max_output
    cdef public double min_output
    cdef double _lower_bounds
    cdef double _upper_bounds

cdef class LogisticParameter(Parameter):
    cdef double _value
    cdef public double max_output
    cdef public double growth_rate
    cdef double _lower_bounds
    cdef double _upper_bounds



from libc.setjmp cimport jmp_buf

cdef extern from "glpk.h":
    ctypedef struct glp_prob:
        pass
    ctypedef struct glp_smcp:
        int msg_lev
        int meth
        double tol_bnd
        double tol_dj
        double tol_piv
        double obj_ll
        double obj_ul
        int it_lim
        int tm_lim
        int presolve
    ctypedef struct glp_mpscp:
        pass

    int GLP_MIN = 1  # minimization
    int GLP_MAX = 2  # maximization

    int GLP_FR = 1  # free (unbounded) variable
    int GLP_LO = 2  # variable with lower bound
    int GLP_UP = 3  # variable with upper bound
    int GLP_DB = 4  # double-bounded variable
    int GLP_FX = 5  # fixed variable

    int GLP_UNDEF = 1  # solution is undefined
    int GLP_FEAS = 2  # solution is feasible
    int GLP_INFEAS = 3  # solution is infeasible
    int GLP_NOFEAS = 4  # no feasible solution exists
    int GLP_OPT = 5  # solution is optimal
    int GLP_UNBND = 6  # solution is unbounded

    int GLP_MSG_OFF = 0  # no output
    int GLP_MSG_ERR = 1  # warning and error messages only
    int GLP_MSG_ON = 2  # normal output
    int GLP_MSG_ALL = 3  # full output
    int GLP_MSG_DBG = 4  # debug output

    int GLP_PRIMAL = 1  # use primal simplex
    int GLP_DUALP = 2  # use dual; if it fails, use primal
    int GLP_DUAL = 3  # use dual simplex

    int GLP_MPS_DECK = 1  # fixed (ancient)
    int GLP_MPS_FILE = 2  # free (modern)

    int GLP_ON = 1
    int GLP_OFF = 0

    glp_prob* glp_create_prob()
    void glp_init_smcp(glp_smcp *parm)
    void glp_erase_prob(glp_prob *P)
    void glp_delete_prob(glp_prob *P)
    void glp_free(void *ptr)
    int glp_free_env()

    int glp_add_rows(glp_prob *P, int nrs)
    int glp_add_cols(glp_prob *P, int ncs)

    void glp_set_mat_row(glp_prob *P, int i, int len, const int ind[], const double val[])
    void glp_set_mat_col(glp_prob *P, int j, int len, const int ind[], const double val[])

    void glp_set_row_bnds(glp_prob *P, int i, int type, double lb, double ub)
    void glp_set_row_name(glp_prob *P, int i, const char *name)

    void glp_set_col_bnds(glp_prob *P, int j, int type, double lb, double ub)
    void glp_set_col_name(glp_prob *P, int j, const char *name)

    void glp_set_obj_coef(glp_prob *P, int j, double coef)

    void glp_set_obj_dir(glp_prob *P, int dir)

    void glp_std_basis(glp_prob *P)
    void glp_adv_basis(glp_prob *P, int flags)
    int glp_simplex(glp_prob *P, const glp_smcp *parm)

    int glp_get_status(glp_prob *P)
    int glp_term_out(int flag)
    void glp_term_hook(int (*func)(void *info, const char *s), void *info)
    void glp_error_hook(void (*func)(void *info), void *info)

    double glp_get_row_prim(glp_prob *P, int i)
    double glp_get_col_prim(glp_prob *P, int j)

    int glp_get_num_rows(glp_prob *P)
    int glp_get_num_cols(glp_prob *P)
    int glp_get_num_nz(glp_prob *P)

    int glp_write_mps(glp_prob *P, int fmt, const glp_mpscp *parm, const char *fname)
    int glp_write_lp(glp_prob *lp, const void *parm, const char *fname)
    int glp_write_prob(glp_prob *P, int flags, const char *fname)

    int glp_get_row_stat(glp_prob *P, int i)
    int glp_get_col_stat(glp_prob *P, int i)
    void glp_set_row_stat(glp_prob *P, int i, int state)
    void glp_set_col_stat(glp_prob *P, int i, int state)



import os
import json
import inspect
import warnings
from IPython.core.display import Javascript, display
from jinja2 import Template
from pywr.core import Node
from pywr.core import Model
from pywr.nodes import NodeMeta
from pywr._component import Component
from pywr.parameters._parameters import get_parameter_from_registry

from .figures import *

# load javascript template for d3 graph
folder = os.path.dirname(__file__)
with open(os.path.join(folder, "draw_graph.js"), "r") as f:
    draw_graph_template = Template(f.read())
with open(os.path.join(folder, "save_graph.js"), "r") as f:
    save_graph_template = Template(f.read())
with open(os.path.join(folder, "graph.css"), "r") as f:
    draw_graph_css = f.read()
with open(os.path.join(folder, "template.html"), "r") as f:
    html_template = Template(f.read())


class PywrSchematic:
    def __init__(
        self, model, width=500, height=400, labels=False, attributes=False, css=None
    ):
        """This object contains methods that allow the graph of a pywr model network to be displayed in a jupyter
        notebook or saved to an html file.

        It also contains a method to save the node positions of a notebook graph back to a pywr model json file. Note
        that this method currently does not work if the object has be instantiated using a model object.

        Parameters
        ----------
        model : pywr.core.Model or json-dict that describes a model
            The model to display
        width : int
            The width of the svg canvas to draw the graph on
        height : int
            The height of the svg canvas to draw the graph on
        labels : bool
            If True, each graph node is labelled with its name. If false, the node names are displayed
            during mouseover events
        attributes : bool
            If True, a table of node attributes is displayed during mouseover events
        css : string
            Stylesheet data to use instead of default
        """
        if isinstance(model, Model):
            self.graph = pywr_model_to_d3_json(model, attributes)
            # TODO update when schema branch is merged
            self.json = None
        else:
            self.graph = pywr_json_to_d3_json(model, attributes)
            if isinstance(model, str):
                with open(model) as d:
                    self.json = json.load(d)
            else:
                self.json = model

        self.height = height
        self.width = width
        self.labels = labels
        self.attributes = attributes

        if css is None:
            self.css = draw_graph_css
        else:
            self.css = css

    def draw_graph(self):
        """Draw pywr schematic graph in a jupyter notebook"""
        js = draw_graph_template.render(
            graph=self.graph,
            width=self.width,
            height=self.height,
            element="element",
            labels=self.labels,
            attributes=self.attributes,
            css=self.css.replace("\n", ""),
        )
        display(Javascript(data=js))

    def save_graph(self, filename, save_unfixed=False, filetype="json"):
        """Save a copy of the model JSON with update schematic positions.

        When run in a jupyter notebook this will trigger a download.

        Parameters
        ----------
        filename: str
            The name of the file to save the output data to.
        save_unfixed: bool
            If True, then all node position are saved to output file. If False, only nodes who have had their position
            fixed in the d3 graph have their positions saved.
        filetype: str
            Should be either 'json' to save the model data with updated node positions to a JSON file or 'csv' to save
            node positions to a csv file.
        """

        if filetype not in ["json", "csv"]:
            warnings.warn(
                f"Output filetype '{filetype}' not recognised. Please use either 'json' or 'csv'</p>",
                stacklevel=2,
            )

        if self.json is None and filetype == "json":
            warnings.warn(
                "Node positions cannot be saved to JSON if PywrSchematic object has been instantiated using "
                "a pywr model object. Please use a JSON file path or model dict instead.",
                stacklevel=2,
            )
        else:
            display(
                Javascript(
                    save_graph_template.render(
                        model_data=json.dumps(self.json),
                        height=self.height,
                        width=self.width,
                        save_unfixed=json.dumps(save_unfixed),
                        filename=json.dumps(filename),
                        filetype=json.dumps(filetype),
                    )
                )
            )

    def to_html(self, filename="model.html", title="Model Schematic"):
        """Save an HTML file of schematic

        Parameters
        ----------
        filename: str
            The name of the html file
        title: str
            The schematic title
        """

        # TODO add option to get node position from graph that has already been drawn in a notebook

        js = draw_graph_template.render(
            graph=self.graph,
            width=self.width,
            height=self.height,
            element=json.dumps(".schematic"),
            labels=self.labels,
            attributes=self.attributes,
            css="",
        )

        html = html_template.render(title=title, css=self.css, d3_script=js)

        with open(filename, "w") as f:
            f.write(html)


def draw_graph(model, width=500, height=400, labels=False, attributes=False, css=None):
    """Display a Pywr model using D3 in Jupyter

    Functionality for creating the d3 graph is now in the PywrSchematic object
    """
    schematic = PywrSchematic(
        model, width=width, height=height, labels=labels, attributes=attributes, css=css
    )
    schematic.draw_graph()


def pywr_model_to_d3_json(model, attributes=False):
    """
    Convert a Pywr graph to a structure d3 can display

    Parameters
    ----------
    model : `pywr.core.Model`
    attributes: bool (default=False)
        If True, attribute data for each node is extract
    """
    nodes = []
    node_names = []
    for node in model.graph.nodes():
        if node.parent is None and node.virtual is False:
            nodes.append(node)
            node_names.append(node.name)

    edges = []
    for edge in model.graph.edges():
        node_source, node_target = edge

        # where a link is to/from a subnode, display a link to the parent instead
        if node_source.parent is not None:
            node_source = node_source.parent
        if node_target.parent is not None:
            node_target = node_target.parent

        if node_source is node_target:
            # link is between two subnodes
            continue

        index_source = node_names.index(node_source.name)
        index_target = node_names.index(node_target.name)
        edges.append({"source": index_source, "target": index_target})

    json_nodes = []
    for n, node in enumerate(nodes):
        node_dict = {"name": node.name}
        classes = []
        cls = node.__class__
        classes.append(cls)
        while True:
            for base in cls.__bases__:
                if issubclass(base, Node) and base is not Node:
                    classes.append(base)
            if classes[-1] is cls:
                break
            else:
                cls = classes[-1]
        classes = classes[::-1]
        node_dict["clss"] = [cls.__name__.lower() for cls in classes]
        try:
            node_dict["position"] = node.position["schematic"]
        except KeyError:
            pass

        if attributes:
            node_dict["attributes"] = get_node_attr(node)

        json_nodes.append(node_dict)

    graph = {"nodes": json_nodes, "links": edges}

    return graph


def get_node_attr(node):
    """
    Returns a dictionary that contains node attributes as strings

    Parameters
    ----------
    node : a pywr node object
    """
    attrs = inspect.getmembers(node, lambda a: not (inspect.isroutine(a)))
    attribute_data = []
    for att in attrs:

        attr_name, attr_val = att
        if attr_name.startswith("_"):
            continue
        attr_type = type(attr_val).__name__

        attrs_to_skip = [
            "component_attrs",
            "components",
            "color",
            "model",
            "input",
            "output",
            "inputs",
            "outputs",
            "sub_domain",
            "sub_output",
            "sublinks",
            "visible",
            "fully_qualified_name",
            "allow_isolated",
        ]
        if not attr_val or attr_name.lower() in attrs_to_skip:
            continue

        if isinstance(attr_val, Component):
            attr_val = attr_val.name
            if not attr_val:
                attr_val = attr_type
            else:
                attr_val = attr_val + " - " + attr_type

        if isinstance(attr_val, list):
            new_vals = []
            for val in attr_val:
                val_name = str(val)
                val_name = val_name.replace("[", "").replace("]", "")
                new_vals.append(val_name)
            attr_val = "".join(new_vals)
        else:
            attr_val = str(attr_val)

        attribute_data.append({"attribute": attr_name, "value": attr_val})

    return attribute_data


def pywr_json_to_d3_json(model, attributes=False):
    """
    Converts a JSON file or a JSON-derived dict into structure that d3js can use.

    Parameters
    ----------
    model : dict or str
        str inputs should be a path to a json file containing the model.
    """

    if isinstance(model, str):
        with open(model) as d:
            model = json.load(d)

    nodes = []
    node_classes = create_node_class_trees()
    for node in model["nodes"]:

        if node["type"].lower() in [
            "annualvirtualstorage",
            "virtualstorage",
            "aggregatednode",
            "aggregatedstorage",
            "seasonalvirtualstorage",
        ]:
            # Do not add virtual nodes to the graph
            continue

        json_node = {"name": node["name"], "clss": node_classes[node["type"].lower()]}
        try:
            json_node["position"] = node["position"]["schematic"]
        except KeyError:
            pass

        if attributes:
            json_node["attributes"] = []
            for name, val in node.items():

                if name == "type":
                    continue

                attr_val = val

                if isinstance(val, dict):
                    try:
                        attr_val = get_parameter_from_registry(val["type"]).__name__
                    except KeyError:
                        pass
                elif val in model["parameters"].keys():
                    param = model["parameters"][val]
                    attr_type = get_parameter_from_registry(param["type"]).__name__
                    attr_val = attr_val + " - " + attr_type
                else:
                    attr_val = str(attr_val)

                attr_dict = {"attribute": name, "value": attr_val}
                json_node["attributes"].append(attr_dict)

        nodes.append(json_node)

    nodes_names = [node["name"] for node in nodes]

    edges = []
    for edge in model["edges"]:
        sourceindex = nodes_names.index(edge[0])
        targetindex = nodes_names.index(edge[1])
        edges.append({"source": sourceindex, "target": targetindex})

    graph = {"nodes": nodes, "links": edges}

    return graph


def create_node_class_trees():
    # create class tree for each node type
    node_class_trees = {}
    for name, cls in NodeMeta.node_registry.items():
        classes = [cls]
        while True:
            for base in cls.__bases__:
                if issubclass(base, Node) and base is not Node:
                    classes.append(base)
            if classes[-1] is cls:
                break
            else:
                cls = classes[-1]
        clss = [cls.__name__.lower() for cls in classes[::-1]]
        node_class_trees[name] = clss
    return node_class_trees



"""Utilities for working with pandas DataFrame objects."""
import pandas
from pandas.tseries.offsets import Tick, DateOffset
from pandas._libs.tslibs.period import IncompatibleFrequency
import os
from .hashes import check_hash


class ResamplingError(Exception):
    def __init__(self, original_dataframe, target_index):
        self.original_dataframe = original_dataframe
        self.target_index = target_index
        message = (
            f'Failed to convert input dataframe with index "{original_dataframe.index}" to model index '
            f'"{target_index}".'
        )
        super().__init__(message)


def align_and_resample_dataframe(df, target_index, resample_func="mean"):
    """Align and resample a DataFrame to the provided index.

    This function attempts to align an input dataframe to a target index. The
     index of the incoming dataframe must be time based and will be cast to a
    `PeriodIndex` (using `to_period()`). Up or down sampling is selected based
     on the relative frequencies of the two indexes. The exact method depends
     also on the type of frequencies (offsets) involved (e.g. `Tick` or
     `DataOffset`). Up-sampling uses forward filling. Whereas down-sampling
     uses the provided `resample_func`.

    Parameters
    ==========

    df : `pandas.DataFrame`
        The input data that needs to be aligned and/or resampled.
    target_index : `pandas.PeriodIndex`
        The target index that the input should be aligned and resampled to match.
    resample_func : str, func
        Function to be used when down-sampling from high frequency data to lower
        frequency.

    """
    # Must resample and align the DataFrame to the model.
    start = target_index[0]
    end = target_index[-1]

    if not isinstance(df.index, pandas.PeriodIndex):
        # Converting to period is sometimes unreliable. E.g. with freq='7D'
        # If the target frequency is passed explicitly this can help, but
        # not all Timestamp frequencies convert to Period frequencies. Therefore,
        # this can not be the default.
        try:
            df = df.to_period()
        except AttributeError:
            df = df.to_period(df.index.freq.freqstr)

    if not isinstance(target_index, pandas.PeriodIndex):
        raise ValueError("Period index expected.")

    model_freq = target_index.freq
    df_freq = df.index.freq

    # Determine how to do the resampling based on the frequency type
    # and whether to do up or down sampling.
    if isinstance(model_freq, Tick):
        # Model is tick based e.g. daily or hourly
        if isinstance(df_freq, Tick):
            # Dataframe is also tick based
            if model_freq >= df_freq:
                # Down sampling (i.e. from high freq to lower model freq)
                df = _down_sample_tick_to_tick(df, target_index, resample_func)
            else:
                df = _up_sample_tick_to_tick(df, target_index)
        else:
            # Dataframe must be offset based
            # Generally this is going to be up sampling to a higher model freq
            df = _resample_date_offset_to_tick(df, target_index, resample_func)

    elif isinstance(model_freq, DateOffset):
        # Model is based on date offsets.
        if isinstance(df_freq, Tick):
            # Dataframe is tick based
            # Generally this is going to be down sampling to a lower model freq
            df = _down_sample_tick_to_date_offset(df, target_index, resample_func)
        else:
            # Dataframe must be offset based
            # Down sampling (i.e. from high freq to lower model freq)
            try:
                df = _resample_date_offset_to_date_offset(
                    df, target_index, resample_func
                )
            except IncompatibleFrequency:
                raise ResamplingError(df, target_index)

    df = df[start:end]

    if not df.index.equals(target_index):
        raise ResamplingError(df, target_index)

    if df.isnull().values.any():
        raise ValueError("Missing values detected after resampling dataframe.")

    return df


def _resample_date_offset_to_date_offset(df, target_index, resample_func):
    """Down sample a date offset data to a target index with date offset."""
    new_df = df.resample(target_index.freq).agg(resample_func)
    new_df = new_df[target_index[0] : target_index[-1]]
    return new_df


def _down_sample_tick_to_date_offset(df, target_index, resample_func):
    """Down sampling tick data to a target index with date offset."""
    new_df = df.resample(target_index.freq).agg(resample_func)
    new_df = new_df[target_index[0] : target_index[-1]]
    return new_df


def _down_sample_tick_to_tick(df, target_index, resample_func):
    """Down sampling tick data to a tick based target index."""
    # First we try to align the higher frequency data to lower the frequency
    start = target_index[0].asfreq(df.index.freq, how="start")
    end = target_index[-1].asfreq(df.index.freq, how="end")
    new_df = df[start:end]
    # Second we re-sample the aligned data
    new_df = new_df.resample(target_index.freq).agg(resample_func)
    return new_df


def _resample_date_offset_to_tick(df, target_index, resample_func):
    """Re-sample a date offset to tick based index."""

    target_tick = target_index.freq

    # Create a copy of the target tick with a period of 1
    target_single_tick = target_tick.base

    # Forward fill the data to the single period tick
    new_df = df.resample(target_single_tick).ffill()

    # Now align to the target index
    start = target_index[0].asfreq(target_single_tick, how="start")
    end = target_index[-1].asfreq(target_single_tick, how="end")
    new_df = new_df[start:end]

    # If the target is multi-frequency
    if target_tick.n > 1:
        new_df = new_df.resample(target_tick).agg(resample_func)
    return new_df


def _up_sample_tick_to_tick(df, target_index):
    """Up sample tick data to a tick based target index."""
    new_df = df.resample(target_index.freq).ffill()
    return new_df


def load_dataframe(model, data):
    column = data.pop("column", None)
    if isinstance(column, list):
        # Cast multiindex to a tuple to ensure .loc works correctly
        column = tuple(column)

    index = data.pop("index", None)
    if isinstance(index, list):
        # Cast multiindex to a tuple to ensure .loc works correctly
        index = tuple(index)

    indexes = data.pop("indexes", None)

    table_ref = data.pop("table", None)
    if table_ref is not None:
        name = table_ref
        df = model.tables[table_ref]
    else:
        name = data.get("url", None)
        df = read_dataframe(model, data)

    # if column is not specified, use the whole dataframe
    if column is not None:
        try:
            df = df[column]
        except KeyError:
            raise KeyError('Column "{}" not found in dataset "{}"'.format(column, name))

    if index is not None:
        try:
            df = df.loc[index]
        except KeyError:
            raise KeyError('Index "{}" not found in dataset "{}"'.format(index, name))

    if indexes is not None:
        try:
            df = df.loc[indexes, :]
        except KeyError:
            raise KeyError('Indexes "{}" not found in dataset "{}"'.format(index, name))

    try:
        if isinstance(df.index, pandas.DatetimeIndex):
            # Only infer freq if one isn't already found.
            # E.g. HDF stores the saved freq, but CSV tends to have None, but infer to Weekly for example
            if df.index.freq is None:
                freq = pandas.infer_freq(df.index)
                if freq is None:
                    raise IndexError(
                        'Failed to identify frequency of dataset "{}"'.format(name)
                    )
                df = df.asfreq(freq)
    except AttributeError:
        # Probably wasn't a pandas dataframe at this point.
        pass

    return df


def read_dataframe(model, data):
    # values reference data in an external file
    url = data.pop("url", None)
    if url is not None:
        if not os.path.isabs(url) and model.path is not None:
            url = os.path.join(model.path, url)
    else:
        # Must be an embedded dataframe
        df_data = data.pop("data", None)

    if url is None and df_data is None:
        raise ValueError('No data specified. Provide a "url" or "data" key.')

    if url is not None:
        # Check hashes if given before reading the data
        checksums = data.pop("checksum", {})
        for algo, hash in checksums.items():
            check_hash(url, hash, algorithm=algo)

        try:
            filetype = data.pop("filetype")
        except KeyError:
            # guess file type based on extension
            if url.endswith((".xls", ".xlsx")):
                filetype = "excel"
            elif url.endswith((".csv", ".gz")):
                filetype = "csv"
            elif url.endswith((".hdf", ".hdf5", ".h5")):
                filetype = "hdf"
            else:
                raise NotImplementedError('Unknown file extension: "{}"'.format(url))
    else:
        if "filetype" in data:
            raise ValueError('"filetype" is only valid when loading data from a URL.')
        if "checksum" in data:
            raise ValueError('"checksum" is only valid when loading data from a URL.')

        filetype = "dict"

    data.pop("comment", None)  # remove kwargs from data before passing to Pandas

    if filetype == "csv":
        df = pandas.read_csv(url, **data)  # automatically decompressed gzipped data!
    elif filetype == "excel":
        df = pandas.read_excel(url, **data)
    elif filetype == "hdf":
        key = data.pop("key", None)
        df = pandas.read_hdf(url, key=key, **data)
    elif filetype == "dict":
        parse_dates = data.pop("parse_dates", False)
        df = pandas.DataFrame.from_dict(df_data, **data)
        if parse_dates:
            df.index = pandas.DatetimeIndex(df.index)

    if df.index.dtype.name == "object" and data.get("parse_dates", False):
        # catch dates that haven't been parsed yet
        raise TypeError(
            'Invalid DataFrame index type "{}" in "{}".'.format(
                df.index.dtype.name, url
            )
        )

    # clean up
    # Assume all keywords are consumed by pandas.read_* functions
    data.clear()

    return df
